import matplotlib.pyplot as plt
from numpy import unique
import networkx as nx
import math
import pandas as pd
from ExpoSeq.tidy_data.clustering import cleaning
import community


def calculate_nodesize(sample_report, region_string, G):
    nodesize = []
    for g in G:
        for i in sample_report[region_string]:
            if i == g:
                node = math.sqrt(sample_report.loc[sample_report[region_string] == i, "cloneFraction"].values[0])
                nodesize.append(node)
                break
    for y, n in enumerate(nodesize):
        nodesize[y] = int(n * 500)
    return nodesize

def clusterSeq(ax, sequencing_report, samples: list,max_ld, min_ld, batch_size, font_settings, second_figure, region_string, label_type = "numbers"):
    sample_report = sequencing_report.loc[sequencing_report["Experiment"].isin(samples)]
    sample_report = sample_report.groupby("Experiment").head(batch_size)
    G = cleaning(sample_report, max_ld, min_ld, region_string)
    # remove nodes that are not connected to any other nodes
    G_deg = G.degree()
    to_remove = [n for (n, deg) in G_deg if deg == 0]
    G.remove_nodes_from(to_remove)


    # Create a gridspec for adding subplots of different sizes
    nodesize = calculate_nodesize(sample_report, region_string, G)

    ## labels
    label_sequences = {}
    node_ids = list(G.nodes())
    n = 0
    for index, g in enumerate(G):
        if n == 9:
            label_sequences[g]= g
            n = 0
        else:
            label_sequences[g] = ""
        n += 1

    label_numbers = {}
    node_ids = list(G.nodes())
    n = 0
    for index, g in enumerate(G):
        if n ==1:
            label_numbers[g] = index
            n = 0
        else:
            n +=1
            label_numbers[g] = index


    partition = community.best_partition(G)
    pos = nx.spring_layout(G)
    nx.draw_networkx_nodes(G, pos, node_size=nodesize,cmap=plt.cm.RdYlBu, node_color=list(partition.values()), ax = ax)
    nx.draw_networkx_edges(G, pos, alpha=0.3, ax = ax)
    if label_type == "numbers":
        nx.draw_networkx_labels(G, pos,
                                labels = label_numbers,
                                alpha = 0.5,
                                font_color="grey",
                                horizontalalignment="right",
                                verticalalignment="bottom", ax = ax)
    elif label_type == "sequence":
        nx.draw_networkx_labels(G, pos,
                                labels = label_sequences,
                                alpha = 0.5,
                                font_color="grey",
                                horizontalalignment="right",
                                verticalalignment="bottom", ax = ax)

    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 22
    ax.set_title(f"Connected components of {samples}", **font_settings)
    font_settings["fontsize"] = original_fontsize
    ax.set_axis_off()

    ## create report

    # Convert each RGBA to nearest named color
   # color_names = [closest_color((r, g, b)) for r, g, b, _ in rgba_colors]
   # color_names = hex_list_to_names(color_names)
    cluster_report = pd.DataFrame([partition]).T
    cluster_report['Sequences'] = cluster_report.index
    cluster_report.reset_index(drop=True, inplace=True)
    cluster_report.rename(columns={0: 'Cluster No.'}, inplace=True)

    return cluster_report

def cluster_single_AG(fig, sequencing_report, antigen, binding_data, max_ld, min_ld, batch_size,region_string, preferred_cmap = "Blues", specific_experiments = False, label_type = "numbers" ):
    assert antigen in binding_data.columns, "it seems like your antigen does not exist in the binding data. Please enter the correct value"
    report_batch = sequencing_report.groupby("Experiment").head(batch_size)
    if specific_experiments != False:
        report_batch = report_batch[report_batch['Experiment'].isin(specific_experiments)]
    else:
        pass

    experiments = report_batch["Experiment"].unique()
    Tot = experiments.shape[0]
    Cols = int(input("How many Columns for the Plots do you want?"))
    Rows = Tot // Cols
    if Tot % Cols != 0:
        Rows += 1
    # Create a Position index
    Position = range(1, Tot + 1)
    plot_number = 0
    cluster_report = pd.DataFrame([])
    for experiment in experiments:
        ax = fig.add_subplot(Rows, Cols, Position[plot_number])
        batch = report_batch[report_batch["Experiment"] == experiment]
        mix = pd.concat([batch, binding_data])
        mix = mix.fillna(0)
        mix = mix.reset_index()
        G = cleaning(mix, max_ld, min_ld, region_string)
        # remove nodes that are not connected to any other nodes
        G_deg = G.degree()
        to_remove = [n for (n, deg) in G_deg if deg == 0]
        G.remove_nodes_from(to_remove)

        nodesize = []
        clone_counts = mix["readCount"]
        color_values = []
        for i in mix.index:
            for g in G:
                if i == g:
                    node = math.sqrt(mix.loc[mix[region_string] == i, "cloneFraction"].values[0])
                    nodesize.append(node)
                    color_values.append(mix.loc[mix[region_string] == i, antigen].values[0])

        for y, n in enumerate(nodesize):
            nodesize[y] = int(n * 500)

        ## labels
        label_sequences = {}
        node_ids = list(G.nodes())
        n = 0
        for index, g in enumerate(G):
            if n == 9:
                label_sequences[g] = g
                n = 0
            else:
                label_sequences[g] = ""
            n += 1

        label_numbers = {}
        node_ids = list(G.nodes())
        n = 0
        for index, g in enumerate(G):
            if n == 1:
                label_numbers[g] = index
                n = 0
            else:
                n += 1
                label_numbers[g] = index

        cm = plt.get_cmap(preferred_cmap)
        vmin = min(color_values)
        vmax = max(color_values)

        partition = community.best_partition(G)
        pos = nx.spring_layout(G)
        nodes = nx.draw_networkx_nodes(G, pos, node_size=nodesize, cmap=cm, node_color=color_values,
                               ax=ax, vmin = vmin, vmax = vmax)
        nx.draw_networkx_edges(G, pos, alpha=0.3)
        if label_type == "numbers":
            nx.draw_networkx_labels(G, pos,
                                    labels=label_numbers,
                                    alpha=0.5,
                                    font_color="grey",
                                    horizontalalignment="right",
                                    verticalalignment="bottom", ax=ax)
        elif label_type == "sequence":
            nx.draw_networkx_labels(G, pos,
                                    labels=label_sequences,
                                    alpha=0.5,
                                    font_color="grey",
                                    horizontalalignment="right",
                                    verticalalignment="bottom", ax=ax)

        cluster_report_inter = pd.DataFrame([partition]).T
        cluster_report_inter[f'Sequences_{experiment}'] = cluster_report_inter.index
        cluster_report_inter.reset_index(drop=True, inplace=True)
        cluster_report_inter.rename(columns={0: f'Clusters_{experiment}_No.'}, inplace=True)
        cluster_report = pd.concat([cluster_report, cluster_report_inter])
        ax.set_title(experiment)
     #   sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=vmin, vmax=vmax))
      #  cbar = plt.colorbar(sm)
        plot_number = plot_number+1
    fig.colorbar(nodes,
                 ax=fig.axes)
    return cluster_report


### is not in use
def cluster_antigens(sequencing_report, sample, antigens, batch_size):
    report_batch = sequencing_report.groupby("Experiment").head(batch_size)
    Tot = len(antigens)
    Cols = int(input("How many Columns for the Plots do you want?"))
    # Compute Rows required
    Rows = Tot // Cols
    #     EDIT for correct number of rows:
    #     If one additional row is necessary -> add one:
    if Tot % Cols != 0:
        Rows += 1
    # Create a Position index
    Position = range(1, Tot + 1)
    plot_number = 0
    fig = plt.figure(1)
    max_colorbar = 0
    batch = report_batch[report_batch["Experiment"] == sample]
    mix = pd.concat([batch, delfia_samples])
    mix = mix.fillna(0)
    mix = mix.reset_index()
    mix = mix.drop("index", axis=1)
    G, degree_sequence = cleaning(mix, max_ld, min_ld, region_string)
    G_deg = G.degree()
    to_remove = [n for (n, deg) in G_deg if deg == 0]
    G.remove_nodes_from(to_remove)
    for antigen in antigens:
        ax = fig.add_subplot(Rows, Cols, Position[plot_number])
        nodesize = []
        clone_counts = mix.readCount
        color_values = []

        for i in range(mix.shape[0]):
            for g in G:
                if i == g:
                    node = math.sqrt(clone_counts.iloc[i])
                    nodesize.append(node)
                    color_values.append(mix[antigen].iloc[i])
        for y, n in enumerate(nodesize):
            if n>0 and n<15:
                nodesize[y] = 15
            if n == 0:
                nodesize[y] = 50

        colormap = plt.get_cmap("Blues")
        vmin = min(color_values)
        vmax = max(color_values)
        if vmax > max_colorbar:
            max_colorbar = vmax
        #   net = networkx.draw_networkx(G, node_size=nodesize, node_color = color_values, cmap = cm,with_labels=False, edgecolors = "Black", ax = ax)
        net = networkx.draw_networkx_nodes(G,
                                           pos=networkx.spring_layout(G),
                                           node_size=nodesize,
                                           node_color=color_values,
                                           cmap=colormap,
                                     edgecolors="Black")
        ax.set_title(antigen)
        #   sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=0, vmax=max_colorbar))
        #  cbar = plt.colorbar(sm)
        plot_number = plot_number + 1
    fig.suptitle("Clustering for sample " + sample)
#    sm = plt.cm.ScalarMappable(cmap=colormap,
 #                              norm=plt.Normalize(vmin = vmin, vmax=max_colorbar))

    #fig.colorbar(ax, col_ax)


















from .mixcr_nils import process_mixcr
from ast import literal_eval
import os
from .loop_collect_reports import load_mixed_files, load_alignment_reports
import pandas as pd
from glob import glob
from ExpoSeq.augment_data.check_reports import check_completeness
import sys
import pickle
from ExpoSeq.augment_data.trimming import trimming
try:
    import tkinter as tk
    from tkinter import filedialog
except:
    pass
import random
import string
import shutil
from ExpoSeq.settings.available_presets import data_list

def get_random_string(length):
    letters = string.ascii_letters  # this will get all the uppercase and lowercase letters
    result_str = ''.join(random.choice(letters) for i in range(length))
    return result_str

def pull_seq_align_repo(testing, testing_value):
    if not testing:
        try:
            print("choose the directory where you store your alignment reports and your sequencing report")
            filenames_dir = filedialog.askdirectory()
            assert os.path.isdir(filenames_dir), "Please enter a valid directory."
            filenames = glob(filenames_dir + "//*")
            
        except:
            filenames_dir = input(
                "Enter the directory where you store the alignment reports and your sequencing report manually.\nYou should have one file for the sequencing report and multiple files for the alignment reports.")
            while True:
                if os.path.isdir(filenames_dir):
                    break
                else:
                    print("Please enter a valid directory.")
                    filenames_dir = input(
                        "Enter the directory where you store the alignment reports and your sequencing report manually.\nYou should have one file for the sequencing report and multiple files for the alignment reports.")
            filenames_dir = os.path.abspath(filenames_dir + "//*")
            filenames = glob(filenames_dir)
    else:
        module_dir = os.path.abspath("")
        filenames_dir = os.path.join(module_dir,
                                    "tests",
                                    "test_files")
        filenames = glob(filenames_dir + "//*")
    for i in filenames:
        if os.path.isdir(i):
            if "alignment_reports" in i:
                alignment_repos = glob(i)
                filenames.extend(alignment_repos)
        sequencing_report, all_alignment_reports = load_mixed_files(filenames, testing, testing_value)
    return sequencing_report, all_alignment_reports

def check_experiment(module_dir, testing):
    if not testing:
        while True:
            experiment = input("How do you want to call your new experiment?")
            if os.path.isdir(os.path.join(module_dir, "my_experiments", experiment)) == True:
                replace = input("The given directory already exists. Do you want to replace it? (Y/n)")
                if replace in ["Y", "y", "n", "N"]:
                    break
                else:
                    print("Please enter another name.")
            else:
                replace = "n"
                break
        if replace in ["Y", "y"]:
            shutil.rmtree(os.path.join(module_dir, "my_experiments", experiment))

    else: 
        experiment = get_random_string(10)
    
    os.mkdir(os.path.join(module_dir,
                    "my_experiments",
                    experiment))
    return experiment

def create_alignment_report(module_dir, experiment):
    try:
        align_repo_path = os.path.join(module_dir,
                                       "my_experiemnts",
                                       experiment,
                                       "alignment_reports",
                                       "*")
        alignment_reports = glob(align_repo_path)
        all_alignment_reports = load_alignment_reports(alignment_reports)
    except:
        all_alignment_reports = pd.DataFrame([])
        print(
            "No alignment reports could be found in " + experiment + ". You will continue without being able to analyze the Alignment Quality.")
    return all_alignment_reports

def method_one(experiment, repo_path, module_dir, testing, paired_end_test = "n"):
    if not testing:
        for i in data_list:
            print(i)
        use_method = input(
            "Per default you will align your data with the following method: milab-human-tcr-dna-multiplex-cdr3.\nPress enter if you want to continue.\nOtherwise type in the method of your choice.\nIt has to be the exact same string which is given on the Mixcr documentation.The available presets are:")
        if use_method == "":
            use_method = "milab-human-tcr-dna-multiplex-cdr3"
        else:
            while True:
                if use_method in data_list:
                    break
                use_method = input("The method you entered is not in the list. Please enter a valid method.")
                if use_method in data_list:
                    break
                else:
                    print("Please enter a valid method.")
    else:
        use_method = ""
    if use_method == "":
        method = "milab-human-tcr-dna-multiplex-cdr3"
    else:
        method = use_method
    if not testing:
        paired_end = input("Do you want to analyze paired_end_sequencing data? (Y/n)")
        
    else: 
        paired_end = 'n'
    while True:
        if paired_end in ["Y", "y", "n", "N"]:
            break
        else:
            print("Please enter a correct value.")
            paired_end = input("Do you want to analyze paired_end_sequencing data? (Y/n)")
    if paired_end.lower() in ["Y", "y"]:
        paired_end = True
    else:
        paired_end = False
    if repo_path == "":
        repo_path = os.path.join(module_dir,
                                "my_experiments",
                                experiment,
                                "sequencing_report.csv")
    process_mixcr(experiment,
                  method=method,
                  testing = testing,
                  paired_end_sequencing=paired_end)

    with open(repo_path, "rb") as f:
        sequencing_report = pd.read_table(f, sep=",")
    all_alignment_reports = create_alignment_report(module_dir, experiment)
    return sequencing_report, all_alignment_reports

def method_two(module_dir, experiment, testing = False, testing_value = None):

    print("Choose the folder which contains the sequencing_report and the alignment reports.")

    sequencing_report, all_alignment_reports = pull_seq_align_repo(testing, testing_value)

    while True:
        if sequencing_report.shape[0] == 0:
            print(
                "Something went wrong. Unfortunately you could not collect the data in the sequencing report.")
            sequencing_report, all_alignment_reports = pull_seq_align_repo(testing, testing_value)
        else:
            break
        #  sys.exit()
    if not testing:
        while True:
            trim_data = input("Do you need to trim your data? (Y/n)")
            if trim_data in ["Y", "y", "n", "N"]:
                break
            else:
                print("Please enter a correct value.")
    else: trim_data = "Y"
    if trim_data.lower() in ["Y", "y"]:
        sequencing_report = trimming(sequencing_report,
                                     divisible_by=3,  # might be possible that columns are different
                                     min_count=1,
                                     new_fraction="cloneFraction")
    else:
        pass
    seq_report_dir = os.path.join(module_dir,
                                  "my_experiments",
                                  experiment,
                                  "sequencing_report.csv")
    sequencing_report.to_csv(seq_report_dir)
    unique_experiments = sequencing_report["Experiment"].unique()
    experiment_dic = {item: item for item in list(unique_experiments)}
    exp_names_dir = os.path.join(module_dir,
                                 "my_experiments",
                                 experiment,
                                 "experiment_names.pickle")
    with open(exp_names_dir, "wb") as f:
        pickle.dump(experiment_dic, f)

    if all_alignment_reports.shape[0] == 0:
        print(
            "No Alignment Reports were uploaded. You will continue the analysis without being able to analyze the Alignment Quality.")
    else:
        try:
            all_alignment_reports = check_completeness(all_alignment_reports, sequencing_report)
        except:
            pass
    exp_name_path = os.path.join(module_dir,
                                 "my_experiments",
                                 experiment,
                                 "experiment_names.pickle")

    unique_experiments = sequencing_report["Experiment"].unique()
    experiment_dic = {item: item for item in list(unique_experiments)}

    with open(exp_name_path, "wb") as f:
        pickle.dump(experiment_dic, f)
    return sequencing_report, all_alignment_reports

def method_three(module_dir, testing ):
    
    if not testing:
        try:
            print("choose the directory with your experiment name and which has the sequencing_report.csv file in its root.")
            while True:
                f = filedialog.askdirectory()
                if os.path.isdir(f):
                    if "sequencing_report.csv" in os.listdir(f):
                        break
                    
                else:
                    print("Please enter a valid path to the directory")
        except:
            f = input("give the path to the directory with the sequencing_report.csv file in its root.")
            f = os.path.abspath(f)
            while True:
                if os.path.isdir(f):
                    if "sequencing_report.csv" in os.listdir(f):
                        break
                else:
                    print("Please enter a valid path to the directory")
        experiment = os.path.basename(f)
        shutil.copytree(f, os.path.join(module_dir, "my_experiments", experiment))
        f = os.path.join(f, "sequencing_report.csv")
    else:
        f = os.path.join(module_dir,
                        "my_experiments",
                        "test_directory",
                        "sequencing_report.csv")
        experiment = "test_directory"
    sequencing_report = pd.read_table(f, sep=",")
    if not testing:
        all_alignment_reports = None
    else:
        all_alignment_reports = None
    return sequencing_report,all_alignment_reports, experiment


def find_file_in_subdirectories(directory, filename):
    for dirpath, dirnames, files in os.walk(directory):
        if filename in files:
            if os.path.basename(dirpath) != "test_directory":
                return os.path.join(dirpath, filename)
    return None

def check_last_exp(module_dir, testing):
    if not testing:
        glob_vars = os.path.join(module_dir,
                                "settings",
                                "global_vars.txt")
        with open(glob_vars, "r") as f:
            data = f.read()
        data = literal_eval(data)

        last_experiment = data["last_experiment"]
        if last_experiment == "":
            if not bool(os.listdir(os.path.join(module_dir, "my_experiments"))):
                repo_path = ""
                print("You have no experiments in your folder.")
            else:
                subdirectory = find_file_in_subdirectories(os.path.join(module_dir, "my_experiments"), "sequencing_report.csv")
                if subdirectory is not None:
                    repo_path = subdirectory
                    dirpath = os.path.dirname(subdirectory)
                    last_experiment = os.path.basename(dirpath)
                else:
                    print("The experiments in my experiment do not contain the sequencing report.\nYou need to upload a new experiment.")
                    repo_path = ""
            
        else:
            repo_path = os.path.join(module_dir,
                                   "my_experiments",
                                    last_experiment,
                                    "sequencing_report.csv")
    else:
        repo_path = os.path.join(module_dir,
                                "my_experiments",
                                "test_directory",
                                "sequencing_report.csv")
        last_experiment = "test_directory"
    return repo_path, last_experiment



def get_continue_analysis_input(last_experiment, testing=False, testing_value = None):
    if not testing:
        while True:
            continue_input = input("Do you want to continue to analyze with " + last_experiment + "? Y/n")
            if continue_input in ["Y", "y", "n", "N"]:
                break
            else:
                print("Please enter a correct value.")
    else:
        continue_input = testing_value
    return continue_input

def get_next_step_input(testing=False, testing_value = None):
    if not testing:
        while True:
            next_step = input("If you want to upload a new experiment press 1.\nIf you want to choose another experiment press 2. ")
            if next_step in ["1", "2"]:
                break
            else:
                print("Please enter a correct value.")
    else:
        next_step = testing_value
    return next_step

def get_choose_method_input(testing, testing_value = None):
    if not testing:
        while True:
            processing = input("If you want to process your fastq files with mixcr press 1.\nIf you want have processed your files externally and want to upload the directory with the Experiment name and the processing data, press 2.")
            if processing in ["1", "2"]:
                break
            else:
                print("Please enter a correct value.")
    else:
        processing = testing_value
    return processing

def get_experiment_name_input(testing = False):
    if not testing:
        experiment_name = input("Enter the name of the experiment you want to analyze")
    else:
        experiment_name = get_random_string(10)
    return experiment_name  

def upload_new_experiment(module_dir, repo_path, testing, testing_value, paired_end_test, experiment_column):
    

    
    while True:
        choose_method = get_choose_method_input(testing, testing_value)
        if choose_method == "1":
            experiment = check_experiment(module_dir, testing)
            
        if choose_method in ["1", "2", "3"]:
            if choose_method == "1":
            
                sequencing_report, all_alignment_reports = method_one(experiment, repo_path,module_dir,testing, paired_end_test )
         #   elif choose_method == "2":
          #      sequencing_report, all_alignment_reports = method_two(module_dir, experiment,testing, experiment_column)
            elif choose_method == "2":
                sequencing_report, all_alignment_reports, experiment = method_three(module_dir, testing)
            break
        else:
            print("Please enter a correct value.")

        

    return sequencing_report, all_alignment_reports, experiment

def choose_existing_experiment(module_dir, testing):
    if not testing:
        while True:
            user_input = get_experiment_name_input()
            spec_exp_name_path = os.path.join(module_dir, "my_experiments", user_input)
            if os.path.isdir(spec_exp_name_path): 
                break
            else:
                print("The experiment name does not exist in my_experiments. Please enter the correct name")
    else:
        user_input = "test_directory"
    return user_input  # experiment



def retrieve_reports(module_dir, experiment, testing):
    if not testing:
        seq_report_path = os.path.join(module_dir, "my_experiments", experiment, "sequencing_report.csv")
        align_repo_path = os.path.join(module_dir,
                            "my_experiments",
                            experiment,
                            "alignment_reports", "*")
    else:
        seq_report_path = os.path.join(module_dir, "my_experiments", "test_directory", "sequencing_report.csv")
        align_repo_path = os.path.join(module_dir, "tests", "test_directory", "all_alignment_reports.pickle")
    if os.path.isfile(seq_report_path):
        with open(seq_report_path, "rb") as f:
            sequencing_report = pd.read_table(f, sep=",")
        try:
            align_repo_path = os.path.join(module_dir,
                                        "my_experiments",
                                        experiment,
                                        "alignment_reports", "*")
            alignment_reports = glob(align_repo_path)
            all_alignment_reports = load_alignment_reports(alignment_reports)
        except:
            all_alignment_reports = pd.DataFrame([])
            print("No alignment reports could be found in " + experiment + ".\nYou will continue without being able to analyze the Alignment Quality.")
    else:
        print("The sequencing report could not be found in " + experiment + ".\nPlease enter an experiment with an existing sequencing_report.csv file or upload new data.")
        sequencing_report = None
        all_alignment_reports = None
        experiment = None

    return sequencing_report, all_alignment_reports, experiment

def process_data_with_last_experiment(module_dir, last_experiment, testing):
    experiment = last_experiment
    sequencing_report, all_alignment_reports, experiment = retrieve_reports(module_dir, experiment, testing)
    return sequencing_report, all_alignment_reports, experiment



def write_last_exp(module_dir, experiment):
    glob_vars = os.path.join(module_dir,
                             "settings",
                             "global_vars.txt")
    with open(glob_vars, "r") as f:
        data = f.read()
    data = literal_eval(data)
    data["last_experiment"] = experiment
    with open(glob_vars, "w") as f:
        f.write(str(data))

def upload(testing=False, continue_analysis = "n", upload_type = "2", choose_exp = '1', paired_end_test = 'n', experiment_column = '1'):
    module_dir = os.getcwd()

    repo_path, last_experiment = check_last_exp( module_dir, testing)
    
    if os.path.isfile(repo_path):
        continue_analysis = get_continue_analysis_input(last_experiment,
                                                        testing,
                                                        continue_analysis)
        if continue_analysis.lower() in ["n", "no"]:
            next_step = get_next_step_input(testing, choose_exp)
            while True:
                if next_step == "1":
                    
                    sequencing_report, all_alignment_reports, experiment = upload_new_experiment(module_dir,
                                                                                                repo_path,
                                                                                                testing,
                                                                                                upload_type,
                                                                                                paired_end_test,
                                                                                                experiment_column)
                elif next_step == "2":
                    experiment = choose_existing_experiment(module_dir,
                                                            testing)
                    sequencing_report, all_alignment_reports, experiment = retrieve_reports(module_dir,
                                                                                            experiment,
                                                                                            testing)
                if isinstance(sequencing_report, pd.DataFrame):
                    break
                else:
                    print("Something went wrong. Please try again.")
                    

        else:
            sequencing_report, all_alignment_reports, experiment = process_data_with_last_experiment(module_dir,
                                                                                                     last_experiment,
                                                                                                     testing)
    else:
        sequencing_report, all_alignment_reports, experiment = upload_new_experiment(module_dir,
                                                                                     repo_path,
                                                                                     testing,
                                                                                     upload_type,
                                                                                     paired_end_test,
                                                                                     experiment_column)
    if testing and experiment != "test_directory":
        shutil.rmtree(os.path.join(module_dir,
                                   "my_experiments",
                                   experiment))
    
    if not testing and experiment != "test_directory":
        write_last_exp(module_dir, experiment)
    return sequencing_report, all_alignment_reports, experiment 



import os

class CreateCommand:
    def __init__(self, module_dir, path_to_mixcr, paired_end_sequencing, experiment, files, java_heap_size, threads = None):
        self.java_heap_size = java_heap_size
        self.module_dir = module_dir
        self.path_to_mixcr = path_to_mixcr
        self.threads = threads
        self.paired_end_sequencing = paired_end_sequencing
        if paired_end_sequencing:
            self.files = [os.path.normpath(j) for i in files for j in i]
            basename = os.path.basename(os.path.splitext(files[0][0])[0])
        else:
            self.files = files
            basename = os.path.basename(os.path.splitext(files[0])[0])
        self.basename = basename
        self.result = os.path.join(self.module_dir,
                                   "temp",
                                   self.basename + ".vdjca")
        self.alignment_path = os.path.normpath(os.path.join(self.module_dir,
                                                            "my_experiments",
                                                            experiment,
                                                            "alignment_reports",
                                                            self.basename + "_AlignmentReport.txt"))
        self.assembly_path = os.path.normpath(os.path.join(self.module_dir,
                                                           "my_experiments",
                                                           experiment,
                                                           "assembly_reports",
                                                           self.basename + "_AssemblyReport.txt"))
        self.clones = os.path.join(self.module_dir,
                                   "my_experiments",
                                   experiment,
                                   "clones_result",
                                   self.basename + "_clones.clns")
        self.clones_base = os.path.dirname(self.clones)
        self.table_tsv = os.path.join(module_dir,
                                     "my_experiments",
                                     experiment,
                                     "tables_mixcr",
                                     self.basename + ".tsv")
        self.mixcr_plots_path = os.path.normpath(os.path.join(self.module_dir,
                                                              "my_experiments",
                                                              experiment,
                                                              "mixcr_plots"))

    def read_aligned_reads(self):
        with open(self.alignment_path, "r") as f:
            alignment = f.readlines()
        for i in alignment:
            if "Successfully aligned reads" in i:
                quality = i.replace("Successfully aligned reads: ", "")
        align_quality = quality.split(" ")[0]
        percent_quality = quality.split(" ")[1]
        align_quality = int(align_quality)
        return align_quality, percent_quality

    def read_clones(self):
        with open(self.assembly_path, "r") as f:
            alignment = f.readlines()
        for i in alignment:
            if "Reads used in clonotypes, percent of total" in i:
                quality = i.replace("Reads used in clonotypes, percent of total: ", "")
        align_quality = quality.split(" ")[0]
        percent_quality = quality.split(" ")[1]
        assembly_quality = int(align_quality)
        return assembly_quality, percent_quality

    def get_free_memory_linux(self):
        with open("/proc/meminfo", "r") as f:
            lines = f.readlines()
        for line in lines:
            if line.startswith("MemAvailable"):
                return int(line.split()[1]) // 1024  # Convert from KB to MB

    def get_free_memory_windows(self):
        try:
            import psutil
            return psutil.virtual_memory().available // (1024 ** 2)  # Convert from Bytes to MB
        except ImportError:
            print("Please install psutil library: pip install psutil")


    def get_free_memory_mac(self):
        cmd = "vm_stat | grep 'Pages free' | awk '{print $3}'"
        pages_free = int(os.popen(cmd).read().replace(".", ""))
        # Typically, a page in macOS is 4096 bytes
        return (pages_free * 4096) // (1024 ** 2)  # Convert from Bytes to MB



    def create_parser(self):
     #   try:
      #      free_memory = self.get_free_memory()
      #      java_heap_size = int(free_memory / 4)  # Using half of the available RAM as an example
       #     print(f"25 % of currently available memory: {java_heap_size}\n")
        #except:
         #   print("automatic detection of memory failed. Processing continues with 1000MB RAM.")
          #  java_heap_size = 1000
        commands = []
        commands.extend(["java", f"-Xms{self.java_heap_size}M", "-jar"]) # enable change of para
        commands.extend([self.path_to_mixcr])
        return commands


    def prepare_align(self, method):
        align_commands = self.create_parser()
        align_commands.extend(["align"])
        print(method)
        align_commands.extend(["--preset", method])
        align_commands.extend(self.files)
        align_commands.extend([self.result])
        align_commands.extend(["--no-warnings"])
        align_commands.extend(["--force-overwrite"])
        if self.threads != None:
            align_commands.extend(["--threads", self.threads])
        align_commands.extend(["--report", self.alignment_path])
        return align_commands

    def prepare_assembly(self):

        assembly_commands = self.create_parser()
        assembly_commands.extend(["assemble"])
       # assembly_commands.extend(["-OseparateByC=true", "-OseparateByV=true","-OseparateByJ=true"])
        assembly_commands.extend([self.result])
        assembly_commands.extend([self.clones])
        assembly_commands.extend(["--force-overwrite"])
        if self.threads != None:
            assembly_commands.extend(["--threads", self.threads])
        assembly_commands.extend(["--report", self.assembly_path])

        return assembly_commands


    def prepare_clones(self):
        clones_commands = self.create_parser()
        clones_commands.extend(["exportClones"])
        clones_commands.extend(["-c IGH"])
        clones_commands.extend([self.clones])
        clones_commands.extend(["--force-overwrite"])
        if self.threads != None:
            clones_commands.extend(["--threads", self.threads])
        clones_commands.extend([self.table_tsv])
        return clones_commands




import numpy as np
import random
import string
import os
import pandas as pd
from Bio.Seq import Seq
import pickle

import math
def create_fractions(num_fractions, top_fraction):
    total_sum = 1
    decrement_factor = math.pow(top_fraction, 1/(num_fractions-1))

    # Create a list of fractions that decrease geometrically
    fractions = [top_fraction * math.pow(decrement_factor, i) for i in range(num_fractions)]
    fractions_sum = sum(fractions)

    # Normalize the fractions so they sum to one
    fractions = [fraction/fractions_sum for fraction in fractions]
    return fractions


def create_sequencing_report(num_experiments = 30, mean_length = 48, stddev_length = 36,panning_rounds = 3, threshold_seq_length = 5 ):
    # Define the stop codons
    stop_codons = ["TAA", "TAG", "TGA"]
    num_experiments = round(num_experiments/panning_rounds)
    sequencing_report = pd.DataFrame([])
    for experiment in range(num_experiments):
        # Define the set of possible nucleotides
        nucleotides = "ATCG"
        sequences = []
        num_sequences = random.randint(1000, 20000)
        # Generate multiple random DNA sequences
        for i in range(num_sequences):
            length = int(np.random.normal(mean_length, stddev_length))
            length = length - (length % 3)  # Round down to the nearest multiple of 3
            while True:
                sequence = ''.join(random.choice(nucleotides) for _ in range(length))
                if not any(sequence[i:i + 3] in stop_codons for i in range(0, len(sequence), 3)):
                    break


            dna_seq = sequence
            sequences.append(dna_seq)
            dna_seq = Seq(dna_seq)
            aa_seq = dna_seq.translate()
        sequences = [string for string in sequences if len(string) >= threshold_seq_length]
        aminoacid_seqs = [str(Seq(string).translate()) for string in sequences]
        num_sequences = len(sequences)
        sample_name = ["sample " + str(experiment)] * num_sequences
        random_float = random.uniform(0, 1) * 0.1
        fractions = create_fractions(num_sequences, random_float)
        random_read = random.randint(1000000, 10000000)
        readsCount = np.array(fractions) * random_read
        readsCount = list(np.round(readsCount).astype(np.uint32))
        length_cdr3 = [len(string) for string in sequences]
        limit = len(aminoacid_seqs)
        cloneid = [i for i in range(1, limit + 1)]
        data = {"Experiment": sample_name,
                "cloneId": cloneid,
                    "aaSeqCDR3": aminoacid_seqs,
                    "nSeqCDR3": sequences,
                    "clonesFraction": fractions,
                    "readCount": readsCount,
                    "lengthOfCDR3": length_cdr3}
        intermediate_frame = pd.DataFrame(data)
        sequencing_report = pd.concat([sequencing_report, intermediate_frame])
        for panning_round in range(panning_rounds-1):
            random_float = random.uniform(0.6, 1)
            fraction_to_sample = 0.5
            num_values_to_sample = int(fraction_to_sample * len(sequences))
            sequences_new = random.sample(sequences, num_values_to_sample)
            aminoacid_seqs_new = [str(Seq(string).translate()) for string in sequences_new]
            num_sequences = len(sequences_new)
            sample_name = ["sample " + str(experiment) + " round " + str(panning_round + 2)] * num_sequences
            random_float = random.uniform(0, 1) * 0.1
            fractions_new = create_fractions(num_sequences, random_float)
            random_read = random.randint(1000000, 10000000)
            readsCount_new = np.array(fractions_new) * random_read
            readsCount_new = list(np.round(readsCount_new).astype(np.uint32))
            length_cdr3_new = [len(string) for string in sequences_new]
            limit = len(aminoacid_seqs_new)
            cloneid = [i for i in range(1, limit + 1)]
            data = {"Experiment": sample_name,
                    "cloneId": cloneid,
                    "aaSeqCDR3": aminoacid_seqs_new,
                    "nSeqCDR3": sequences_new,
                    "clonesFraction": fractions_new,
                    "readCount": readsCount_new,
                    "lengthOfCDR3": length_cdr3_new}
            intermediate_frame = pd.DataFrame(data)
            sequencing_report = pd.concat([sequencing_report, intermediate_frame])
    return sequencing_report



def create_binding_report(sequencing_report, num_antigen):
    threshold_aa_seq = 5
    sampled_df = sequencing_report[~sequencing_report.duplicated(subset='aaSeqCDR3', keep=False)]
    sampled_df = sampled_df.groupby("Experiment").sample(frac = 0.01)
    sampled_df = sampled_df[sampled_df['aaSeqCDR3'].apply(len) >= threshold_aa_seq]
    aaSeq = sampled_df["aaSeqCDR3"].to_list()
    fractions = sampled_df.clonesFraction
    binding_values = []
    for i in fractions:
        random_binding = random.randint(1000000, 10000000000)
        binding_value = i * random_binding
        binding_values.append(binding_value)
    num_splits = num_antigen
    # generate random indices where the splits should occur
    split_indices = np.sort(np.random.choice(range(1, len(binding_values)), num_splits, replace=False))
    # split the list into random parts
    split_lists = [(binding_values[i:j], aaSeq[i:j]) for i, j in zip([0]+split_indices, split_indices+[len(binding_values)])]
    for i in range(len(split_lists)):
        part_binding = split_lists[i][0]
        part_aaSeq = split_lists[i][1]
        data = {"aaSeqCDR3": part_aaSeq, "Epitope " + str(i): part_binding}
        intermediate_binding = pd.DataFrame(data)
        if i == 0:
            binding_data = intermediate_binding
        else:
            binding_data = pd.merge(binding_data, intermediate_binding,how = "left",on = "aaSeqCDR3")
    
    return binding_data


def create_test_dataset(module_dir):
    sequencing_report = create_sequencing_report()
    binding_data = create_binding_report(sequencing_report, 5)
    binding_data_filepath = os.path.join(module_dir, "test_data", "test_files", "binding_data.csv")
    assert os.path.isfile(binding_data_filepath), "binding data filepath does not exist"
    binding_data.to_csv(binding_data_filepath)
    sequencing_report_filepath = os.path.join(module_dir, "test_data", "test_files", "sequencing_report.csv")
    assert os.path.isfile(sequencing_report_filepath), "sequencing report filepath does not exist"
    sequencing_report.to_csv(sequencing_report_filepath)
    unique_experiments = sequencing_report["Experiment"].unique()
    experiment_dic = {item: item for item in list(unique_experiments)}
    exp_names_dir = os.path.join(module_dir, "test_data", "test_files", "experiment_names.pickle")
    assert os.path.isfile(exp_names_dir), "experiment names filepath does not exist"
    with open(exp_names_dir, "wb") as f:
        pickle.dump(experiment_dic, f)
        



import math
def best_layout(num_subplots):
    """
    Determine the best layout for a given number of subplots.

    Parameters:
        - num_subplots: The total number of subplots

    Returns:
        - rows, columns: Best number of rows and columns for the layout
    """

    # If it's a perfect square
    if math.sqrt(num_subplots).is_integer():
        return int(math.sqrt(num_subplots)), int(math.sqrt(num_subplots))

    # If not a perfect square
    for i in range(int(math.sqrt(num_subplots)), 0, -1):
        if num_subplots % i == 0:
            return i, num_subplots // i

    raise ValueError("Unable to determine layout")


import os


def check_dirs(module_dir, experiment):
    if not os.path.isdir(os.path.join(module_dir, "my_experiments")):
        os.mkdir(os.path.join(module_dir, "my_experiments"))
    if not os.path.isdir(os.path.join(module_dir, "my_experiments", experiment)):
        os.mkdir(os.path.join(module_dir, "my_experiments", experiment))      
    if not os.path.isdir(os.path.join(module_dir, "my_experiments", experiment, "alignment_reports")):
        os.mkdir(os.path.join(module_dir, "my_experiments", experiment, "alignment_reports"))
    if not os.path.isdir(os.path.join(module_dir, "temp")):
        os.mkdir(os.path.join(module_dir, "temp"))   
    if not os.path.isdir(os.path.join(module_dir, "my_experiments", experiment, "assembly_reports")):
        os.mkdir(os.path.join(module_dir, "my_experiments", experiment, "assembly_reports"))
    if not os.path.isdir(os.path.join(module_dir, "my_experiments", experiment, "clones_result")):
        os.mkdir(os.path.join(module_dir, "my_experiments", experiment, "clones_result"))
    if not os.path.isdir(os.path.join(module_dir,
                                     "my_experiments",
                                     experiment,
                                     "tables_mixcr")):
        os.mkdir(os.path.join(module_dir, "my_experiments", experiment, "tables_mixcr"))


import pandas as pd

def add_fraction(sequencing_report,
                group = 'Experiment',
                percentage_from = 'readCount',
                new_column_name = 'cloneFraction'):
    pd.options.mode.chained_assignment = None # otherwise error will be raised
    sequencing_report[new_column_name] = sequencing_report.groupby([group])[percentage_from].transform(lambda x: x / x.sum())
        # raise error if name exists
    return sequencing_report


def mapFunc(sequencing_report, column, func, column_name):
    mapped_column = sequencing_report[column].apply(func)
    sequencing_report[column_name] = mapped_column
    return sequencing_report




from ExpoSeq.augment_data.create_mixcr_commands import CreateCommand
from ExpoSeq.augment_data.collect_fastqs import CollectFastq
import subprocess
import os
import pandas as pd
import pickle
import argparse
from ExpoSeq.settings.check_dirs import check_dirs
        
def mixcr_(path_to_mixcr, experiment_name,  path_to_forward, path_to_backward = None, method = "ampliseq-tcrb-plus-full-length", threads = 1, java_heap_size = None):
    print(path_to_mixcr)
    print(path_to_forward)
    print(path_to_backward)
    print(method)
    print(threads)
    print(java_heap_size)
    module_dir = os.path.abspath("")
    assert os.path.isfile(path_to_mixcr), "File path to mixcr.jar file is incorrect."
    assert path_to_forward != path_to_backward, "Directory to forward and backward files needs to be different"
    assert os.path.isdir(path_to_forward), "The given directory for the forward files does not exist."
    assert subprocess.run(["java", "-jar",path_to_mixcr,"--version"]), "The given path to mixcr is not correct. Or you have not installed mixcr correctly."
    assert type(experiment_name) == str, "Please enter a string as experiment name"
  #  assert os.path.isfile(module_dir, "my_experiments", experiment_name) == False, "The experiment name already exists, please enter another one."
    check_dirs(module_dir, experiment_name)
    if java_heap_size == None:
        java_heap_size = 1000
    if path_to_backward == None:
        paired_end_sequencing = False
    else:
        paired_end_sequencing = True

    Collect = CollectFastq(paired_end_sequencing)
    
    Collect.module_dir = module_dir

    Collect.get_files(cmd = True,
                      path_to_files = path_to_forward,
                      path_to_backward = path_to_backward)
    files = Collect.paired
    print(files)
    if paired_end_sequencing:
        print("Found pairs for the forward and reverse files")
        Collect.call_pairs()

            
    sequencing_report = pd.DataFrame([])
    for filename in files:
        Commands = CreateCommand(module_dir, path_to_mixcr, paired_end_sequencing, experiment_name, filename)
        print("go")
        subprocess.run(Commands.prepare_align(method))
        skip_sample = False
        
        align_quality, percent_quality = Commands.read_aligned_reads()
        if align_quality < 1000 and align_quality >10:
            print(f"Only {align_quality} reads could be aligned ({percent_quality}).")
        if align_quality < 10:
            print("Too less reads were aligned. Please choose the correct mixcr method.")
            skip_sample = True
        if skip_sample == False:
            subprocess.run(Commands.prepare_assembly())
            assembly_qual, realtive_assembly_quality = Commands.read_clones()
            if assembly_qual < 1000 and realtive_assembly_quality > 10:
                print(f"Only {assembly_qual} reads could be assembled ({realtive_assembly_quality}).")
            if assembly_qual < 10:
                print("Too less reads were assembled. Please choose the correct ")
                skip_sample = True
            if skip_sample == False:
                subprocess.run(Commands.prepare_clones())
                basename = Commands.basename

                for filename in os.listdir(os.path.join(module_dir, "temp")):
                    if filename.endswith(".tsv"):
                        clones_sample = pd.read_table(os.path.join(module_dir, "temp", filename))
                clones_sample["Experiment"] = basename
                sequencing_report = pd.concat([sequencing_report, clones_sample])

                for file in os.listdir(os.path.join(module_dir, "temp")):
                    os.remove(os.path.join(module_dir,
                                        "temp",
                                        file))
     #   except:
      #      print(f"File {filename} could not be processed. Please check if you have chosen the correct mixcr method or verify that the sequencing was successful.")

    sequencing_report.to_csv(os.path.join(module_dir, experiment_name, 'sequencing_report.csv'))
    if sequencing_report.shape[0] == 0:
        raise Exception(
            "The processing failed for all of your files. Most likely because you chose the wrong mixcr method.")

    unique_experiments = sequencing_report["Experiment"].unique()
    experiment_dic = {item: item for item in list(unique_experiments)}
    exp_names_dir = os.path.join(module_dir,
                                    "my_experiments",
                                    experiment_name,
                                    "experiment_names.pickle")
    with open(exp_names_dir, "wb") as f:
        pickle.dump(experiment_dic, f)
    print(
    "The output of your processing is a sequencing_report which contains the sequences and their sample names of all fastq files. Further, a folder with the alignment reports for each fastq file was created. You can have a look under ~/my_experiments/YOUR_EXPERIMENT_NAME")



parser = argparse.ArgumentParser(description='Processing')
parser.add_argument('path_to_mixcr', type=str, help="Path to the mixcr jar file")
parser.add_argument('experiment_name', type=str, help="Name of the experiment")
parser.add_argument('--path_to_forward',type = str, help='Directory to the fastq files with the forward reads')
parser.add_argument('--path_to_backward',type = str,default=None, help='Directory to the fastq files with the backward reads')
parser.add_argument('--threads', default=1, type=int, help='Number of threads to use')
parser.add_argument("--method", default="milab-human-tcr-dna-multiplex-cdr3", type=str, help="Method to use for the alignment")
parser.add_argument("--java_heap_size", default="1000", type=int, help="Memory allocation for the processing in MB")

args = parser.parse_args()
path_to_mixcr = args.path_to_mixcr
experiment_name = args.experiment_name
path_to_forward = args.path_to_forward
path_to_backward = args.path_to_backward
threads = args.threads
method = args.method
java_heap_size = args.java_heap_size

# Call the function
mixcr_(path_to_mixcr, experiment_name, path_to_forward, path_to_backward,  threads, method, java_heap_size)
#PlotManager(experiment = experiment_name)


import pandas as pd
import editdistance
import numpy as np

def create_distance_matrix(aa):
    num_sequences = len(aa)
    distance_matrix = np.zeros((num_sequences, num_sequences))  # initialize with zeros

    for i in range(num_sequences):
        for j in range(i+1, num_sequences):
            lev_distance = editdistance.distance(aa[i], aa[j])
            distance_matrix[i, j] = lev_distance
            distance_matrix[j, i] = lev_distance  # The distance matrix is symmetric

    return distance_matrix

def get_clustered_sequences(aa, max_cluster_dist):
    clustered_sequences = set()

    for i in range(len(aa)):
        for j in range(i+1, len(aa)):
            if editdistance.distance(aa[i], aa[j]) < max_cluster_dist:
                clustered_sequences.add(aa[i])
                clustered_sequences.add(aa[j])

    return list(clustered_sequences)


def label_bind_seqs(mix, region_string, aa_clustered, antigens):
    values = mix.loc[mix[region_string].isin(aa_clustered), antigens]
    #values = pd.DataFrame(values, columns = ["binding"])
    filtered_df = values[antigens][values[antigens] > 1].dropna(how='all')
    indices = filtered_df.index.tolist()
    key_sequences = []
    key_values = []
    for i in indices:
        key_sequences.append(mix.iloc[i][region_string])
        max_value = max(mix.iloc[i][column] for column in antigens)
        key_values.append(max_value)
    seq_val = pd.DataFrame([key_sequences, key_values]).T
    seq_val.columns = ["binding_seqs", "binding_values"]
    seq_val = seq_val.sort_values(by = "binding_values", ascending = False)
    seq_val.reset_index(inplace = True)
    binding_seqs = seq_val.binding_seqs.to_list()
    for label in range(len(aa_clustered)):
        label_str = aa_clustered[label]
        if label_str in binding_seqs:
            pos_label = binding_seqs.index(label_str)
            label_str = "(" + str(pos_label + 1) + ") " + label_str
            binding_seqs[pos_label] = label_str
        aa_clustered[label] = label_str
    return aa_clustered, binding_seqs, seq_val

def sort_binding_seqs(binding_seqs, seq_val, ascending=True):

    if ascending:
        zipped = list(zip(binding_seqs, seq_val.binding_values))

        # Sort by binding_values
        zipped_sorted = sorted(zipped, key=lambda x: x[1])

        # Unzip the sorted list
        binding_seqs_sorted, binding_values_sorted = zip(*zipped_sorted)
    else:
        binding_seqs_sorted = binding_seqs
        binding_values_sorted = seq_val.binding_values.to_list()
    return binding_seqs_sorted, binding_values_sorted


import editdistance
import pandas as pd

# create an empty dictionary to store the summarized clone fractions

def cleaning_all():
    data = {}
    for sample in samples:
        report = sequencing_report[sequencing_report["Experiment"] == sample]
        report = report[report[region_string].str.len() >= length_filter]
        sequences = report[region_string].to_list()
        
# display the summarized dataframe
def cleaning(sequencing_report,max_levenshtein_distance, samples, length_filter,region_string, batch):
    data = {}
    for sample in samples:
        report = sequencing_report[sequencing_report["Experiment"] == sample]
        report = report[report[region_string].str.len() >= length_filter]
        sequences = report[region_string].head(batch).to_list()
        clone_fractions = report["cloneFraction"].head(batch).to_list()
        max_distance = max_levenshtein_distance
        summary = {}
        total_clone_fraction = 0

        # iterate over the sequences in the dataframe
        for i in range(len(sequences)):
            sequence = sequences[i]
            clone_fraction = clone_fractions[i]
            total_clone_fraction += clone_fraction
            # check if the sequence is already in the summary dictionary
            added = False
            for key in summary.keys():
                if editdistance.distance(sequence, key) <= max_distance:
                    summary[key] += clone_fraction
                    added = True
                    break
            # if the sequence is not in the summary dictionary, add it
            if not added:
                summary[sequence] = clone_fraction
        count_data = {}
        # convert the summary dictionary to a pandas dataframe
        summary = sorted(summary.items(), key=lambda x: x[1], reverse=True)
        first_20 = dict(summary[:20])
        intermediate_df = pd.DataFrame.from_dict(first_20, orient='index')
        data[sample] = intermediate_df
    all_samples = pd.concat(data, axis=1)
    all_samples = all_samples.fillna(0)
    all_samples = all_samples.rename(columns={col: col[0].replace('/sample', "") for col in all_samples.columns})
    all_samples.columns = samples
    return all_samples








from gensim.models import Word2Vec
from gensim.models.keyedvectors import KeyedVectors
from Bio import SeqIO
import numpy as np

from Bio.Align import substitution_matrices
# Load the BLOSUM62 matrix from Biopython
blosum = np.array(substitution_matrices.load("BLOSUM62"), dtype=int)
blosum = blosum[:20, :20]
# Define a function to generate training examples

fraction = 0.01

# group your data by sample
grouped = sequencing_report.groupby(['Experiment', 'aaSeqCDR3'])

# loop over each sample and sample the desired fraction of sequences
samples = []
for name, group in grouped:

    sampled_group = sequencing_report.groupby("Experiment")["aaSeqCDR3"].head(100).sample(frac=1)
    # append the sampled sequences to the list of samples
    samples.append(sampled_group)

# concatenate the samples into a new DataFrame
sampled_data = pd.concat(samples)



def generate_training_examples(seq, window=5):
    examples = []
    collect_all = []
    for i, target in enumerate(seq):
        for j in range(max(0, i- window), min(len(seq), i + window + 1)):
            if i != j:
                context = seq[j]
                collect_all.append((target, context))
    sampling = random.sample(collect_all, 10)
    examples.extend(sampling)


    return examples


# Define a function to generate embeddings
def generate_embeddings(sequences_full, embedding_size=100, window=6, iter=5):
    # Read the protein sequences from a FASTA file

    aminoacids = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']
    # Generate the training examples
    training_examples = []
    for seq in sequences_full:
        seq_vec = []
        for aa in seq:
            if aa in aminoacids:
                seq_vec.append(list(blosum[aminoacids.index(aa)]))
        training_examples.extend(generate_training_examples(seq_vec, window=window))

    batch = []
    for tup in training_examples:
        for word in tup:
            batch.append(word)
    # Train the Word2Vec model
    model = Word2Vec(vector_size=embedding_size, window=window, min_count=1, sg=1)
    model.build_vocab(batch)
    model.train(batch, total_examples=model.corpus_count, epochs=model.epochs)

    # Save the embeddings to a file
    model.wv.save_word2vec_format('protvec.txt')

    # Load the embeddings into a KeyedVectors object
    embeddings = KeyedVectors.load_word2vec_format('protvec.txt', binary=False)

    return embeddings


import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
from ExpoSeq.tidy_data.heatmaps import tidy_jaccard, tidy_sorensen, tidy_morosita_horn, tidy_heatmap_share

import numpy as np


class MorositaHornMatrix:
    def __init__(self,ax, sequencing_report, region_string, samples,protein, font_settings):
        self.ax = ax
        self.samples = samples
        self.matrix = self.cleaningPlot( samples, sequencing_report, protein, region_string, )
       # self.createPlot()

        self.font_settings = font_settings
        self.func_type = {
        "Transparency": ["alpha", [0.01, 1]],
        "Line Width": ["linewidth",[0.01, 1]],
        "annotation size": ["annot_kws",[1, 10]],
        "Color Map": ["cmap", ["virdis", "plasma", "inferno", "magma", "cividis", "BuPu", "PuRd", "jet", "turbo", "hot", "magma", ]],
        "Line Color": ["line_color"]
            }
        self.button_type = {
        "Transparency": "CustomScale",
            "Line Width": "CustomScale",
            "Annotation size": "CustomScale",
            "cmap": "optionlist",
            "Line Color": "SimpleTextField"
        }
        
        
                
    def cleaningPlot(self, samples, sequencing_report, protein, region_string, ):
        unique_sequences, unique_experiments = tidy_morosita_horn.cleaning_data(sequencing_report,
                                                                                region_string,
                                                                                protein = protein,
                                                                                specific_experiments = self.samples,
                                                                                )
        matrix, unique_sequences, unique_experiments = tidy_morosita_horn.morosita_horn_matrix(unique_sequences,
                                                                                        unique_experiments)
        unique_experiments = list(unique_experiments)
        matrix = matrix.sort_index(axis=1)
        matrix = matrix.sort_index(axis = 0)
        return matrix
    
    def createPlot(self,linewidth = 0,line_color = "white", annotate_cells = False, annot_kws = 1, fmt = ".2f", cmap = 'inferno', orientation = "vertical", spacing = 'proportional'):
        annot_kws = {"size": annot_kws}
        unique_experiments = list(unique_experiments)
        matrix = matrix.sort_index(axis=1)
        matrix = matrix.sort_index(axis = 0)
        matplotlib.use('Qt5Agg')
        #colorbar_sets = {**colorbar_settings,
             #               **{'label': "Degree of Identity"}}
        self.morosita_horn_matrix = sns.heatmap(matrix,
                                                ax = self.ax,
                                                linewidths=linewidth,
                                                linecolor = line_color,
                                                annot = annotate_cells,
                                                annot_kws=annot_kws,
                                                fmt = fmt,
                                                cmap= cmap,
                                                orientation = orientation,
                                                spacing = spacing)


        

    def add_style(self, highlight_specific_pos):
        plt.xticks(ticks = np.arange(0.5, len(self.samples) + 0.5 ,1),
                labels = self.samples,
                rotation = 45,
                ha = 'right',
                size = 12) # create a function which finds the perfect size based on counts of xlabels
        plt.yticks(ticks = np.arange(0.5, len(self.samples) + 0.5 ,1),
                labels = self.samples,
                rotation = 0,
                va='center',
                size=12)
        original_fontsize = self.font_settings["fontsize"]
        self.font_settings["fontsize"] = 22
        title = "Identity between samples based on Morosita Horn index"
        self.ax.set_title(title,pad = 12, **self.font_settings)
        self.font_settings["fontsize"] = original_fontsize






def plot_heatmap(sequencing_report, protein, heatmap, ax, colorbar_settings, font_settings, annotate_cells,region_of_interest, specific_experiments = False):

    if heatmap == "morosita_horn":
        unique_sequences, unique_experiments = tidy_morosita_horn.cleaning_data(sequencing_report,
                                                                                region_of_interest,
                                                                                protein = protein,
                                                                                specific_experiments = specific_experiments,
                                                                                )
        matrix, unique_sequences, unique_experiments = tidy_morosita_horn.morosita_horn_matrix(unique_sequences,
                                                                                               unique_experiments)
        title = "Identity between samples based on Morosita Horn index"
    if heatmap == "jaccard":
        matrix, unique_sequences, unique_experiments = tidy_jaccard.cleaning_jaccard(sequencing_report,
                                                                                     region_of_interest,
                                                                                     protein=protein,
                                                                                     specific_experiments = specific_experiments,
                                                                                     )
        title = "Identity between samples based on Jaccard index"
    if heatmap == "sorensen":
        matrix, unique_sequences, unique_experiments = tidy_sorensen.heatmap_sorensen(sequencing_report,
                                                                                      protein,
                                                                                      region_of_interest,
                                                                                      
                                                                                      specific_experiments = specific_experiments)
        title = "Identity between samples based on Sorensen index"
    if heatmap == "relative":
        matrix, unique_sequences, unique_experiments = tidy_heatmap_share.heatmap_share(sequencing_report,
                                                                                        region_of_interest,
                                                                                        protein=protein,
                                                                                        specific_experiments = specific_experiments)
        title = "Identity between samples based on relative coefficient"

    unique_experiments = list(unique_experiments)
    matrix = matrix.sort_index(axis=1)
    matrix = matrix.sort_index(axis = 0)
    matplotlib.use('Qt5Agg')
    colorbar_sets = {**colorbar_settings,
                         **{'label': "Degree of Identity"}}
    sns.heatmap(matrix,
                ax = ax,
                cbar_kws = colorbar_sets,
                annot = annotate_cells,
                annot_kws={"size": 6},
                fmt = ".2f")

    plt.xticks(ticks = np.arange(0.5, len(unique_experiments) + 0.5 ,1),
               labels = unique_experiments,
               rotation = 45,
               ha = 'right',
               size = 12) # create a function which finds the perfect size based on counts of xlabels
    plt.yticks(ticks = np.arange(0.5, len(unique_experiments) + 0.5 ,1),
               labels = unique_experiments,
               rotation = 0,
               va='center',
               size=12)
    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 22
    ax.set_title(title,pad = 12, **font_settings)
    font_settings["fontsize"] = original_fontsize
    return matrix   









import pandas as pd

def read_matrix(threshold, path_matrix):
    matrix = pd.read_excel(path_matrix, index_col = 0)
    # Create the dictionary with column names and the corresponding row indexes
    result = {}
    for col in matrix.columns:
        # Get the row indexes where the value is greater than threshold
        indexes = matrix[matrix[col] > threshold].index.tolist()
        if indexes:  # Only add to the dictionary if there's at least one index
            result[col] = indexes
    return result


import matplotlib.pyplot as plt
from ExpoSeq.tidy_data.barplot import cleaning_data
import numpy as np
import matplotlib


def barplot(ax, all_alignment_reports, sequencing_report_all,font_settings, legend_settings, apply_log):
    matplotlib.use("Qt5Agg")
    boxplot_data_frame = cleaning_data(all_alignment_reports,
                                       sequencing_report_all)
    ax.bar(boxplot_data_frame.Experiment,
           np.array(boxplot_data_frame.tot_sequenced_reads).astype(np.float32),
           label="Total Sequenced Reads",
            color="orange")
    ax.bar(boxplot_data_frame.Experiment,
            np.array(boxplot_data_frame.Aligned_Reads).astype(np.float32),
            label="Aligned Reads",
            color="royalblue")
    plt.xticks(rotation=45,
              ha = 'right',
              size=12)

    ax.legend(**legend_settings)
    plt.ylabel("Reads Count", **font_settings)
    plt.xlabel("Sample", **font_settings)

    if apply_log == True:
        plt.yscale("log")
    plt.tight_layout()
    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 22
    ax.set_title("Alignment Quality of the analyzed samples",pad = 12, **font_settings)
    font_settings["fontsize"] = original_fontsize


   # matplotlib.use("Qt5Agg")
   # boxplot_data_frame = cleaning_data(all_alignment_reports, sequencing_report_all)
   # ax.bar(boxplot_data_frame.Experiment, np.array(boxplot_data_frame.tot_sequenced_reads).astype(np.float32),label="Total Sequenced Reads",color="orange", np.array(boxplot_data_frame.Aligned_Reads).astype(np.float32),
   # label="Aligned Reads",
   # color="royalblue")
    #ax.set_xticks(rotation=45, ha = 'right', size=5)
    #params_legend = openParams("USQ_plot_legend_params.txt")
    #ax.legend(**params_legend)
    #plot_style = openParams('plot_style.txt')
    #ax.set_ylabel("Reads Count", **plot_style)
    #ax.set_xlabel("Sample", **plot_style)
    #if apply_log == True:
     #   ax.set_yscale("log")
    #fig.tight_layout()



from numpy import array
def trimming(sequencing_report, divisible_by = 3, min_count = 3, new_fraction = "cloneFraction"):
    sequencing_report = sequencing_report[(sequencing_report["lengthOfCDR3"] % divisible_by) == 0]
    sequencing_report = sequencing_report[(sequencing_report["readCount"] > min_count)]
    new_column = sequencing_report['readCount'] / sequencing_report.groupby('Experiment')['readCount'].transform('sum')
    sequencing_report[new_fraction] = array(new_column)
    return sequencing_report


#from augment_data.uploader import upload
#from .augment_data.uploader import upload
from ExpoSeq.augment_data.uploader import upload
import os
import shutil
from ExpoSeq.full_pipe import run_pipeline
import pkg_resources

def run_initial_tests():
    pkg_dir = pkg_resources.resource_filename("ExpoSeq", "")
    module_dir = os.path.abspath("")
    if not os.path.isdir(os.path.join(module_dir, "my_experiments")):
        os.mkdir(os.path.join(module_dir, "my_experiments"))
    if not os.path.isdir(os.path.join(module_dir, "temp")):
        os.mkdir(os.path.join(module_dir, "temp"))
    if not os.path.isdir(os.path.join(module_dir, "my_experiments", "test_directory")):
        new_dir = os.path.join(module_dir, "my_experiments", "test_directory")
        os.mkdir(new_dir)
        source_seq_report = os.path.join(pkg_dir, "test_data","test_files", "sequencing_report.csv")
     #   source_align_report = os.path.join(pkg_dir, "test_data", "all_alignment_reports.pickle")
        source_experiments = os.path.join(pkg_dir, "test_data","test_files", "experiment_names.pickle")
        assert shutil.copy2(source_seq_report, new_dir), "could not copy sequencing report"
     #   shutil.copy2(source_align_report, new_dir)
        shutil.copy2(source_experiments, new_dir)
    else:
        pass
    
    #upload(testing=True, continue_analysis = "n", upload_type = "2", choose_exp = '1', paired_end_test = 'n', experiment_column = '1')
    assert upload(testing=True, continue_analysis = "n", upload_type = "1", choose_exp = '1', paired_end_test = 'n', experiment_column = '1'), "test 1 with upload_type 1,choose_exp = 1 and negative analysis continuation failed"
    print("test 1 passed")
    assert upload(testing=True, continue_analysis = "n", upload_type = "1", choose_exp = '2', paired_end_test = 'n', experiment_column = '1'), "test 2 with upload_type 1,choose_exp = 2 and negative analysis continuation failed"
    print("test 2 passed")
    assert upload(testing=True, continue_analysis = "n", upload_type = "3", choose_exp = '1', paired_end_test = 'n', experiment_column = '1'), "test 3 with upload_type 3,choose_exp = 1 and negative analysis continuation failed"	
    print("test 3 passed")
    assert upload(testing=True, continue_analysis = "Y", upload_type = "2", choose_exp = '1', paired_end_test = 'n', experiment_column = '1'), "test 4 with upload_type 2,choose_exp = 1 and positive analysis continuation failed"
    print("test 4 passed")
    
    print(os.path.abspath('run.py'))
    run_pipeline(test = True)
    print("all tests passed")

    
run_initial_tests()


import pandas as pd

def cleaning(sequencing_report, sample,region, region_string):

    sample = sequencing_report[sequencing_report["Experiment"] == sample]
    local_report = sample[["Experiment", "cloneFraction", region_string]]
    aminoacids = "ACDEFGHIKLMNPQRSTVWY"

    sequences = local_report[local_report[region_string].astype(str).str.len() >= region[1]][region_string]
    max_length = local_report[region_string].str.len().max()
    if not region[1] <= max_length:
        raise ValueError("you upper region limit is above the longest sequence. That is not possible. Please reduce it.")
    compDict = {aa: max_length*[0] for aa in aminoacids}
    for seq in sequences:
        for aa_position in range(len(seq)):
            aminoacid = seq[aa_position]
            if aminoacid == '*':
                pass
            else:
                compDict[aminoacid][aa_position] += 1
    aa_distribution = pd.DataFrame.from_dict(compDict)
    aa_distribution = aa_distribution.divide(aa_distribution.sum(axis=1), axis=0)
    aa_distribution = aa_distribution[(aa_distribution.index >= region[0]) & (aa_distribution.index <= region[1])]
    return aa_distribution


import matplotlib.pyplot as plt
import pandas as pd
from sgt.SGT import SGT
from sklearn.decomposition import PCA
import numpy as np
from sklearn.manifold import TSNE


def cluster_toxins_tsne(fig, sequencing_report, sample, toxins, binding_data, toxin_names, pca_components, perplexity, iterations_tsne, font_settings, colorbar_settings, extra_figure, region_of_interest):
    report = sequencing_report[sequencing_report["Experiment"] == sample]
    batch = report.groupby("Experiment").head(1000)
    if extra_figure == True:
        ax = fig.gca()
        fig2 = plt.figure()
        ax2 = fig2.gca()
    else:
        ax = fig.add_subplot(221)
        ax2 = fig.add_subplot(222)
    mix = batch.merge(binding_data, on = region_of_interest, how = "left")
    mix = pd.concat([batch, binding_data])
    mix = mix.fillna(0)

    kds = mix[toxins].max(axis = 1)
    ids = mix[toxins].idxmax(axis = 1)
    corpus = pd.DataFrame([list(kds), mix[region_of_interest].map(list)]).T
    sgt = SGT(kappa = 1, lengthsensitive = False)
    sequences = mix[region_of_interest].map(list)
    sequences_list = list(sequences)
    sgtembedding_df = sgt.fit_transform(corpus=sequences_list)
    pca = PCA(n_components=pca_components)
    pca.fit(sgtembedding_df)
    X = pca.transform(sgtembedding_df)
    print("Explained variance after reducing to " + str(pca_components) + " dimensions:" + str(np.sum(pca.explained_variance_ratio_).tolist()))
    tsne = TSNE(n_components=2,
                verbose=0,
                perplexity=perplexity,
                n_iter=iterations_tsne)
    tsne_results = tsne.fit_transform(X)
    tsne_results = pd.DataFrame(tsne_results,
                                columns = [["tsne1", "tsne2"]])
    x_cor = list(tsne_results.tsne1.iloc[:, 0])
    y_cor = list(tsne_results.tsne2.iloc[:, 0])

    tsne_results = pd.DataFrame(tsne_results,
                                columns=[["tsne1", "tsne2"]])
    aminoacids = mix[region_of_interest].to_list()
    experiments_batch = mix["Experiment"]
    unique_experiments_num = pd.factorize(experiments_batch)[0]
    tsne_results["experiments_factorized"] = list(unique_experiments_num)
    tsne_results["experiments_string"] = list(experiments_batch)
    tsne_results["binding"] = list(kds)
    tsne_results["sequences"] = list(aminoacids)
    tsne_results['sequence_id'] = pd.Series(range(1170))
    tsne_plot = ax.scatter(tsne_results.tsne1,
                           tsne_results.tsne2,

                           c = tsne_results.binding)
    if toxin_names == True:
        for i, txt in enumerate(list(ids)):
            if list(kds)[i] > 0:
                ax.annotate(txt, (x_cor[i], y_cor[i]))
    else:
        pass
    plt.colorbar(tsne_plot, **colorbar_settings)
    ax.set_xlabel("t-SNE1", **font_settings)
    ax.set_ylabel("t-SNE2", **font_settings)
    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 22

    font_settings["fontsize"] = original_fontsize
    fig.tight_layout()


    ax2.scatter(tsne_results.tsne1, tsne_results.tsne2, alpha = 0.0)
    ax2.set_xlabel('t-SNE 1', **font_settings)
    ax2.set_ylabel('t-SNE 2', **font_settings)
    n = 0
    for j, row in tsne_results.iterrows():
        if row["binding"] > 1:
                ax2.text(row['tsne1'], row['tsne2'], row['sequence_id'], fontsize=10, weight = "bold")

        else:
            if n == 6:
                n = 0
                ax2.text(row['tsne1'], row['tsne2'], row['sequence_id'], fontsize=8)
        n += 1

    fig.suptitle("Sequence Embedding on t-SNE Space for Sample " + sample, pad= 12, **font_settings)

    return tsne_results


#from gensim.models import Word2Vec
#import numpy as np

# Define CDR3 sequences
#cdr3_sequences = ['CASSLGAEQYF', 'CASSLGAEQYF']

# Define ProtVec model parameters
#embedding_size = 100
#window_size = 15
#min_word_count = 1
#workers = 8

# Train ProtVec model on CDR3 sequences
#cdr3_tokens = [list(seq) for seq in cdr3_sequences]
#protvec_model = Word2Vec(cdr3_tokens, size=embedding_size, window=window_size, min_count=min_word_count, workers=workers, sg=1)

# Generate ProtVec embeddings for CDR3 sequences
#cdr3_embeddings = np.array([protvec_model.wv[token] for token in cdr3_tokens])

# Print embeddings
#print(cdr3_embeddings)

















import re

def input_pattern():
    filepattern = input()
    return filepattern



def pattern_generator(filepattern): # works only if input is single digit
    filepattern_seperate = list(filepattern)
    local_pattern = ""
    local_pattern_more_digits = ""
    for letter in range(0, len(filepattern_seperate)):
        if filepattern_seperate[letter].isnumeric() != True:
            if filepattern_seperate[letter].isupper():
                local_pattern = local_pattern + '[A-Z]'
                local_pattern_more_digits = local_pattern_more_digits + '[A-Z]'
            elif filepattern_seperate[letter].islower():
                local_pattern = local_pattern + '[a-z]'
                local_pattern_more_digits = local_pattern_more_digits + '[a-z]'
            elif filepattern_seperate[letter].isspace():
                local_pattern = local_pattern + '\s'
                local_pattern_more_digits = local_pattern_more_digits + '\s'
            else:
                local_pattern = local_pattern + '[' + filepattern_seperate[letter] + ']'
                local_pattern_more_digits = local_pattern_more_digits + '[' + filepattern_seperate[letter] + ']'
        else:
            local_pattern_more_digits = local_pattern_more_digits + '\d+'
            local_pattern = local_pattern + '[1-9]'


    return local_pattern, local_pattern_more_digits





def grouping_files(local_pattern, local_pattern_more_digits, filenames): # will not work if you have three numbers in your filepattern on different places and your second number has two digits but your last number has one digit
    used_filenames = []
    grouped_filenames = []
    for index_filename in range(0, len(filenames)):
        subgroup = []
        if not bool(filenames[index_filename] in used_filenames):
            local_pattern_single = re.search(local_pattern,
                                             filenames[index_filename])
            local_pattern_multiple = re.search(local_pattern_more_digits,
                                               filenames[index_filename])
            if local_pattern_single is None:
               # local_pattern_multiple = re.search(local_pattern_more_digits,
                #                                   filenames[index_filename])
                local_pattern_single = local_pattern
                local_pattern_multiple = local_pattern_multiple.group()
            else:
                local_pattern_single = local_pattern_single.group()
                local_pattern_multiple = local_pattern_multiple.group()
            for filepath in filenames:
                if local_pattern_single == local_pattern_multiple:
                    if filepath not in used_filenames and local_pattern_single in filepath:
                        subgroup.append(filepath)
                        used_filenames.append(filepath)
                else:
                    if filepath not in used_filenames and local_pattern_multiple in filepath:
                        subgroup.append(filepath)
                        used_filenames.append(filepath)
            if bool(subgroup) != False:
                grouped_filenames.append(subgroup)
    return grouped_filenames


def find_exp_name(filename): # assumes that company appends uniqueCDR3_Exp_UniqueCDR3 to filenames so that they can be seperated by _
    list_filename = filename.split('\\')
    last_item_list = list_filename[-1]
    local_split = last_item_list.split("_")
    del local_split[len(local_split) - 3:]
    experiment_name = '_'.join(local_split)
    return experiment_name



def cleaning_data(sequencing_report, samples):
    sub_table = sequencing_report.loc[sequencing_report["Experiment"].isin(samples)]
    apply_fun = lambda x: list(x)
    sub_table = sub_table.groupby("Experiment")[['nSeqCDR3', 'cloneFraction', 'readCount']].agg(apply_fun).reset_index()
    return sub_table




def print_instructions():
    instructions = '''

-----------------------
General Instructions
-----------------------
All plots can be called with 'plot' followed by the name of the plot, separated by a dot.
To analyze a specific sample or investigate possible options, use 'help()' and 'plot' followed by the name of the plot.

-----------------------
Plot Commands
-----------------------
- plot.alignment_quality(): 
    Creates a figure showing the Overall Reads per sample and the number of aligned reads.

- plot.aa_distribution(): 
    Returns a plot showing the amino acid distribution in the given sequence range.

- plot.rarefraction_curves(): 
    Shows the rarefraction curves for the given samples.

- plot.logoPlot_single(): 
    A logo Plot showing the composition of amino acids per position.

- plot.logoPlot_multiple(): 
    A logo Plot showing the composition of amino acids per position for multiple samples.

- plot.lengthDistribution_single(): 
    Shows the length distribution of the given sample.

- plot.relative_abundance_multi(): 
    Shows the relative abundance of the given samples.

- plot.lengthDistribution_multi(): 
    Shows the length distribution of the given samples.

- plot.rel_seq_abundance(): 
    Shows a bar plot of the frequencies of the most abundant sequences with optional Levenshtein distance.

- plot.basic_cluster(): 
    Clusters a batch of sequences based on a threshold for the Levenshtein distance.

- plot.cluster_one_AG(): 
    Clusters sequences based on Levenshtein distance and shows binding data against a specific antigen.

- plot.tsne_cluster_AG(): 
    Embeds sequences in a vector space, reduces dimensions, and clusters them. Also plots binding data.

- plot.embedding_tsne(): 
    Transforms sequences into a vector space and uses PCA and t-SNE for dimensionality reduction.

- plot.morosita_horn(): 
    Returns a matrix of the identity between your samples based on the Morosita Horn Index.

- plot.jaccard(): 
    Returns a matrix of the identity between your samples based on the Jaccard Index.

- plot.sorensen(): 
    Returns a matrix of the identity between your samples based on the Sorensen Index.

- plot.relative(): 
    Returns a matrix of the identity between your samples based on the Relative Index.

- plot.levenshtein_dendrogram(): 
    Returns a dendrogram of the sequences based on the Levenshtein distance.

- plot.dendro_bind(): 
    Shows a dendrogram based on Levenshtein distance and a bar plot with binding data.

- plot.MSA_design(): 
    Creates a multiple sequence alignment and offers an interface for sequence design.

-----------------------
Non-Plot Commands
-----------------------
- plot.add_binding_data(): 
    Allows you to add more binding data to the pipeline.

- plot.change_preferred_antigen(): 
    Sets the antigen you'd like to analyze in the plots.

- plot.change_preferred_sample(): 
    Sets the sample you'd like to analyze in the plots.

- plot.change_preferred_region(): 
    Sets the region you'd like to analyze in the plots.

- plot.change_filter(): 
    Adjusts the thresholds for filtering sequences in your analysis.

- plot.discard_samples(): 
    Allows you to remove samples from the analysis.

- plot.merge_bind_seq_report(): 
    Merges binding data with NGS data.

- plot.print_antigens(): 
    Shows the available antigen names in your data.

- plot.print_samples(): 
    Shows the available samples in your data.

- plot.save(): 
    Allows you to save a plot.

- plot.change_experiment_names(): 
    Allows you to change the names of the samples.
'''
    print(instructions)


from tkinter import filedialog
import os
from glob import glob



class CollectFastq():
    def __init__(self, paired_end_sequencing):
        self.forward = []
        self.backward = []
        self.paired = []
        self.paired_end_sequencing = paired_end_sequencing
        
    def add_fastq_files(self):
        filenames = []
        while True:
            try:
                while True:
                    path_to_files = filedialog.askdirectory()
                    if os.path.isdir(path_to_files) == True:
                        break
                    else:
                        pass

            except:
                while True:
                    path_to_files = input("copy and paste the path to your directory")
                    if os.path.isdir(os.path.abspath(path_to_files)) == True:
                        break
                    else:
                        print("Please enter a valid directory. Dont add a \ to the end of the directory path")
            path_to_files = os.path.abspath(path_to_files)
            filenames.extend(glob(os.path.join(path_to_files, "*.fastq*")))
            if len(filenames) == 0:
                user_choice = input(
                    "You have not chosen a directory with fastq files.\n If you want to try again press 1.\n If you want to cancel the analysis press 2")
                if user_choice == str(1):
                    read_more_files = "Y"
                else:
                    read_more_files = "N"
            else:
                print("These are the files you chose:")
                for i in filenames:
                    print(os.path.basename(i))
                while True:
                    read_more_files = input("do you want to add another folder? (Y/n)")
                    if read_more_files in ["Y" "y", "n", "N"]:
                        break
                    else:
                        pass
            if read_more_files == "Y" or read_more_files == "y":
                continue
            else:
                break
        sorted(filenames)
        return filenames


    def get_filenames(self):
        while True:
            filenames = self.add_fastq_files()
            if len(filenames) == 0:
                print("You have not added any fastq files the preprocessing is cancelled now.")
            else:
                break
        return filenames
    def get_filename(self):
        try:
            while True:
                path_to_file = filedialog.askopenfilename()
                if os.path.isfile(path_to_file) == True:
                    break
                else:
                    pass

        except:
            while True:
                path_to_file = input("copy and paste the path to your directory")
                if os.path.isfile(os.path.abspath(path_to_file)) == True:
                    break
                else:
                    print("Please enter a valid directory. Dont add a \ to the end of the directory path")
        return path_to_file
    def __len__(self, file_list):
        return len(file_list)

    def get_pairs(self):
        if self.__len__(self.forward) != self.__len__(self.backward):
            print("Files matching aborted. Your have not collected the same number of files for the forward and backward reads.")
        else:
            best_pair = (None, None)
            # Pairwise comparison of filenames
            for i, file1 in enumerate(self.forward):
                for j, file2 in enumerate(self.backward):
                    with open(file1, "r") as one:
                        one_first = one.readline().strip()
                        one_sub_first = one_first.rfind(":")
                        one_first_sub = one_first[one_sub_first+1:]
                    with open(file2, "r") as two:
                        two_first = two.readline().strip()
                        two_sub_first = two_first.rfind(":")
                        two_first_sub = two_first[two_sub_first+1:]
                    if one_first_sub == two_first_sub:
                        best_pair = [file1, file2]

                if not best_pair:
                    print(f"Could not find match for {file1}")
                else:
                    self.paired.append(best_pair)
    def get_files(self,  cmd = False, path_to_forward = None, path_to_backward = None):
        if cmd == False: 
            print("Choose the directory where you store the fastq files with the forward reads or single end sequencing data. \nIf you want to continue with paired end sequencing data make sure that you store your reverse reads in a seperate folder. \nFurther make sure your chosen directory does not contain fastq files from other experiments.")
            self.forward = self.get_filenames()
            if self.paired_end_sequencing:
                print("Now choose the directory where you store the fastq files with the backward reads.")
                self.backward = self.get_filenames()
                self.get_pairs()
            else:
                self.paired = self.forward
        else:
            path_to_forward= os.path.abspath(path_to_forward)
            self.forward.extend(glob(os.path.join(path_to_forward, "*.fastq*")))
            if self.paired_end_sequencing == True:
                path_to_backward = os.path.abspath(path_to_backward)
                self.backward.extend(glob(os.path.join(path_to_backward, "*.fastq*")))
                self.get_pairs()
            else:
                self.paired = self.forward

                
 

    def call_pairs(self):
         if self.paired_end_sequencing:
             for i in self.paired:
                 print("You will process " + i[0] + " with " + i[1])

    def manually_match_pairs(self):
        self.paired = []
        self.backward = []
        self.forward = []
        while True:
            print("Choose the file of your fastq file with the forward reads.")
            forward_file = self.get_filename()
            self.forward.append(forward_file)
            print("Choose the file of your fastq file with the backward reads")
            backward_file = self.get_filename()
            self.backward.append(backward_file)
            self.paired.append([forward_file, backward_file])
            read_more_files = input("do you want to add more files? (Y/n)")
            if read_more_files in ["Y" "y", "n", "N"]:
                break
            else:
                pass


import os
import sys
from glob import glob
import subprocess
from ast import literal_eval
import pickle
import pandas as pd
from ExpoSeq.augment_data.trimming import trimming
import pkg_resources
try:
    import tkinter as tk
    from tkinter import filedialog
except:
    pass
def add_fastq_files():
    filenames = []
    while True:
        try:
            while True:
                path_to_files = filedialog.askdirectory()
                if os.path.isdir(path_to_files) == True:
                    break
                else:
                    pass
            
        except:
            while True:
                path_to_files = input("copy and paste the path to your directory")
                if os.path.isdir(os.path.abspath(path_to_files)) == True:
                    break
                else:
                    print("Please enter a valid directory. Dont add a \ to the end of the directory path")
        path_to_files = os.path.abspath(path_to_files)
        filenames.extend(glob(os.path.join(path_to_files, "*.fastq")))
        if len(filenames) == 0:
            user_choice = input("You have not chosen a directory with fastq files. If you want to try again press 1. If you want to cancel the analysis press 2")
            if user_choice == str(1):
                read_more_files = "Y"
            else:
                read_more_files = "N"
        else:
            print("These are the files you chose:")
            for i in filenames:
                print(os.path.basename(i))
            while True:
                read_more_files = input("do you want to add another folder? (Y/n)")
                if read_more_files in ["Y" "y", "n", "N"]:
                    break
                else:
                    pass
        if read_more_files == "Y" or read_more_files == "y":
            continue
        else:
            break
    sorted(filenames)

    return filenames


def process_mixcr(experiment, method, testing, paired_end_sequencing):
    pkg_path = pkg_resources.resource_filename("ExpoSeq", "")
    module_dir = os.path.abspath("")
    
    if not testing:
        print("Choose the directory where you store your fastq files")
        while True:
            filenames = add_fastq_files()
            if len(filenames) == 0:
                print("You have not added any fastq files the preprocessing is cancelled now.")
            else:
                break
        if not os.path.isdir(os.path.join(module_dir, "my_experiments", experiment, "alignment_reports")):
            os.mkdir(os.path.join(module_dir, "my_experiments", experiment, "alignment_reports"))

    else:

        filenames = glob(os.path.join(pkg_path,"test_data","test_files", "*.fastq"))
        print(filenames)
        experiment = "test_directory"
    
    settings_dir = os.path.join(module_dir,
                                "settings",
                                "global_vars.txt")
    with open(settings_dir) as f:
        data = f.read()
    data = literal_eval(data)
    path_to_mixcr = data["mixcr_path"]
    if path_to_mixcr == "":
        while True:
            print("choose the mixcr java file with the file chooser")
            try:
                path_to_mixcr = filedialog.askopenfilename()
            except:
                path_to_mixcr = input("copy and paste the path to the mixcr jar file here.")
                path_to_mixcr = os.path.normpath(path_to_mixcr)
            if testing == True:
                break

            
            try:
                print("Test Mixcr")
                subprocess.run(["java",
                                    "-jar",
                                    path_to_mixcr,
                                    "--version"
                                    ])
                confirmation = True
            except:
                print("Apparently you have not choosen the right path to the mixcr .jar file. Please try again")
                confirmation = False
            if confirmation == True:
                path_to_mixcr = os.path.abspath(path_to_mixcr)
                data["mixcr_path"] = path_to_mixcr
                with open(settings_dir, "w") as f:
                    f.write(str(data))
                break
    else:
        pass
    #kAligner2_4.0
    #nebnext-human-bcr-cdr3
    #milab-human-tcr-dna-multiplex-cdr3


    sequencing_report = pd.DataFrame([])
    if paired_end_sequencing == True:
        filenames_unique_reverse = [unique for unique in filenames if "2.fastq" in unique]
        filenames_unique_forward = [unique for unique in filenames if "1.fastq" in unique]
        combined_filenames = []
        for i in filenames_unique_reverse:
            file_one = i
            basename = os.path.basename(os.path.splitext(file_one)[0])
            basename = basename[:len(basename) - 2]
            file_two = ""
            for j in filenames_unique_forward:
                if basename in filenames_unique_forward:
                    file_two = j
                    break
            if file_two == "":
                print("you are missing the the reverse strand for " + basename)
            else:
                fastq_files = file_one + " " + file_two
                combined_filenames.append(fastq_files)
        if len(combined_filenames) == 0:
            print("the system couldnt find any forward reverse matches. Please check your given directory or continue with single-end analysis.")
            return
    else:
        combined_filenames = filenames
    
    for filename in combined_filenames:
        basename = os.path.basename(os.path.splitext(filename)[0])
        basename = basename[:len(basename) - 2]
        filename_base = basename
        output_temp_dir = os.path.join(module_dir, "temp", basename)
        subprocess.run(["java",
                       "-jar",
                       path_to_mixcr,
                       "analzye",
                       method,
                       filename,
                       output_temp_dir])

        clones_sample = pd.read_table(os.path.join(module_dir,
                                                   "temp",
                                                   basename + ".clones_IGH.tsv"))
        clones_sample = clones_sample[["cloneId",
                                        "readCount",
                                        "readFraction",
                                        "nSeqFR1",
                                        "minQualFR1",
                                        "nSeqCDR1",
                                        "minQualCDR1",
                                        "nSeqFR2",
                                        "minQualFR2",
                                        "nSeqCDR2",
                                        "minQualCDR2",
                                        "nSeqFR3",
                                        "minQualFR3",
                                        "nSeqFR4",
                                        "minQualFR4",
                                        "nSeqCDR3",
                                        "aaSeqCDR3",
                                        "minQualCDR3",
                                        "lengthOfCDR3",
                                        "meanQualCDR3"]]
        new_fractions = clones_sample.groupby("nSeqCDR3")["readFraction"].sum().reset_index()
        clones_sample = clones_sample.drop_duplicates(subset=["nSeqCDR3"], keep="first")
        clones_sample = new_fractions.merge(clones_sample,
                                            how = "left",
                                            on = "nSeqCDR3")
        clones_sample = clones_sample.sort_values(by="cloneId")
        clones_sample = clones_sample.reset_index()
        clones_sample = clones_sample.drop(columns = ["readFraction_y", "index"])
        clones_sample = clones_sample.rename(columns={"readFraction_x": "cloneFraction"})
        clones_sample["Experiment"] = filename_base
        clones_sample = trimming(clones_sample,
                                 divisible_by = 3,
                                 min_count = 3,
                                 new_fraction = "clonesFraction")
        sequencing_report = pd.concat([sequencing_report, clones_sample])
        files_to_remove = os.listdir(os.path.join(module_dir, "temp"))
        for file in files_to_remove:
            os.remove(os.path.join(module_dir,
                                   "temp",
                                   file))
    if not testing:
        report_dir = os.path.join(module_dir,
                                  "my_experiments",
                                  experiment,
                                  "sequencing_report.csv")
        sequencing_report.to_csv(report_dir)
        data["last_experiment"] = experiment
        glob_vars_dir = os.path.join(module_dir,
                                     "settings",
                                     "global_vars.txt")
        with open(glob_vars_dir, "w") as f:
            f.write(str(data))
        unique_experiments = sequencing_report["Experiment"].unique()
        experiment_dic = {item: item for item in list(unique_experiments)}
        exp_names_dir = os.path.join(module_dir,
                                     "my_experiments",
                                     experiment,
                                     "experiment_names.pickle")
        with open(exp_names_dir, "wb") as f:
            pickle.dump(experiment_dic, f)
    print("The output of your processing is a sequencing_report which contains the sequences and their sample names of all fastq files. Further, a folder with the alignment reports for each fastq file was created. You can have a look under ~/my_experiments/YOUR_EXPERIMENT_NAME")




import numpy as np
import pandas as pd
from sgt.SGT import SGT
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE


def tidy_embed(sequencing_report, batch_size, list_experiments, pca_components, perplexity ,iterations_tsne, region_string):
    report_batch = sequencing_report.groupby("Experiment").head(batch_size)
    selected_rows = report_batch.loc[report_batch["Experiment"].isin(list_experiments)]
    sgt = SGT(kappa=1,
              lengthsensitive=True)
    sequences = selected_rows[region_string].map(list)
    sequences_list = list(sequences)
    sgtembedding_df = sgt.fit_transform(corpus=sequences_list)
    pca = PCA(n_components=pca_components)
    pca.fit(sgtembedding_df)
    X = pca.transform(sgtembedding_df)
    print(np.sum(pca.explained_variance_ratio_))
    tsne = TSNE(n_components=2,
                verbose=0,
                perplexity=perplexity,
                n_iter=iterations_tsne)
    tsne_results = tsne.fit_transform(X)
    tsne_results = pd.DataFrame(tsne_results,
                                columns=[["tsne1", "tsne2"]])
    aminoacids = selected_rows["aaSeqCDR3"].to_list()
    experiments_batch = selected_rows.Experiment
    unique_experiments_num = pd.factorize(experiments_batch)[0]
    tsne_results["experiments_factorized"] = list(unique_experiments_num)
    tsne_results["experiments_string"] = list(experiments_batch)
    return tsne_results, aminoacids, experiments_batch


import numpy as np
import matplotlib.pyplot as plt
from ExpoSeq.plots.layout_finder import best_layout

def length_distribution_single(fig,ax, sequencing_report, sample, font_settings, region_string):
    batch = sequencing_report[sequencing_report["Experiment"] == sample]
    length = batch[region_string].str.len()
    unique_length, counts_length = np.unique(np.array(length)
                                                , return_counts = True)
    ax.bar(unique_length, counts_length)  # Or whatever you want in the subplot
    #ax.set_xticks(range(0, max_length + 1, 1), range(0, max_length + 1, 1))
    ax.title.set_text(sample)
    ax.title.set_size(18)
    ax.set_ylabel("Read Count",
                    **font_settings)  # Y label
    ax.set_xlabel('Read Length',
                **font_settings)  # X label

    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 20
    plt.title("Length Distribution of " + sample,
              pad=12,
              **font_settings)
    font_settings["fontsize"] = original_fontsize



def length_distribution_multi(fig, sequencing_report, samples, font_settings, region_string, test_version = False,):
    if samples == "all":
        unique_experiments = sequencing_report["Experiment"].unique()
        unique_experiments = np.sort(unique_experiments)
    else:
        sequencing_report = sequencing_report[sequencing_report['Experiment'].isin(samples)]
        unique_experiments = sequencing_report["Experiment"].unique()
        unique_experiments = np.sort(unique_experiments)
    Tot = unique_experiments.shape[0]
    Rows, Cols = best_layout(Tot)
    Position = range(1, Tot + 1)
    n = 0

   # fig = plt.figure(1, constrained_layout=True)
    for experiment in unique_experiments:
        batch = sequencing_report[sequencing_report["Experiment"] == experiment]
        length = batch[region_string].str.len()
        unique_length, counts_length = np.unique(np.array(length)
                                                 , return_counts = True)
            # add every single subplot to the figure with a for loop
        ax = fig.add_subplot(Rows, Cols, Position[n])
        ax.bar(unique_length, counts_length)  # Or whatever you want in the subplot
       # ax.set_xticks(range(0, max_length + 1, 1), range(0, max_length + 1, 1))
        ax.title.set_text(experiment)
        ax.title.set_size(10)

        adapted_fontsize = 10 - int(Cols) + 2
        font_settings["fontsize"] = adapted_fontsize
        ax.set_ylabel("Read Count",
                      **font_settings)  # Y label
        ax.set_xlabel('Read Length',
                      **font_settings)  # X label

        n += 1
    fig.suptitle("Length Distribution of given Samples")







 #   plt.show()
   # plt.bar(unique_length, counts_length)
   # p_values = counts_length/np.sum(counts_length)
    #normalized on one:
   # plt.bar(unique_length, counts_length/np.max(counts_length))
    #p-values
  #  plt.bar(unique_length, p_values)


from Bio.Align.Applications import ClustalwCommandline
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from ExpoSeq.design.msa_design import ExtendedSequenceManipulator
import os
import pandas as pd



class MSA:
    def __init__(self, region_string, SequencingReport,Settings, module_dir):
        self.region_string = region_string
        self.SeqReport = SequencingReport
        self.Settings = Settings
        self.seqs_path = os.path.join(module_dir, "temp", "sequences.fasta")
        self.outfile = os.path.join(module_dir, "temp", "aligned_clustal.fasta")
        
    
    def create_fasta(self, sequences):
        records = [SeqRecord(Seq(row["aaSeqall"]), id=f"seq{i+1}", description='') for i, row in sequences.iterrows()]
        SeqIO.write(records, self.seqs_path, "fasta")

    #def is_allowed(self,s):
     #   allowed_characters = 'ACDEFGHIKLMNPQRSTVWYZ-end_*'
      #  return all(c in allowed_characters for c in s)

    def get_top_sequences(self, batch_size, samples):
        self.SeqReport.filter_longest_sequence()
        sequencing_report = self.SeqReport.sequencing_report


        # Custom function to check if all characters in a string are allowed

        sequencing_report = sequencing_report[sequencing_report['Experiment'].isin(samples)]

      #  sequencing_report = sequencing_report[sequencing_report['aaSeqall'].apply(self.is_allowed)]
    #    sequencing_report = sequencing_report[~sequencing_report['aaSeqall'].str.contains('\*')]
        sequencing_report = sequencing_report.groupby("Experiment").head(batch_size)
        sequences = sequencing_report["aaSeqall"]
        sequences = pd.DataFrame(sequences.values, columns = ["aaSeqall"])
        return sequences

    def get_clustalw_path(self):
        global_params = self.Settings.read_global_vars()
        clustalw_path = global_params["clustalw_path"]
        try:
            import tkinter as tk
            from tkinter import filedialog
            if clustalw_path == '':
                print("Please choose the path to the clustalw exe with the filechooser. If you don't have it, you can download it from http://www.clustal.org/download/current/")
                clustalw_path = filedialog.askopenfilename()
                self.Settings.change_global_vars("clustalw_path", clustalw_path)
        except:
            print("Tk not installed. Please install it to use this functionality and visualize the MSA")
        return clustalw_path
    def run_MSA(self, batch_size, samples):
        path_clustalw = self.get_clustalw_path()
        if not path_clustalw == '' and os.path.isfile(path_clustalw):
            try:
                os.remove(self.outfile)
            except:
                pass
            sequences = self.get_top_sequences(batch_size, samples)
            self.create_fasta(sequences)
            clustalw_cline = ClustalwCommandline(path_clustalw, infile=self.seqs_path, outfile=self.outfile)
            stdout, stderr = clustalw_cline()
            example_seq = "YOUR_SEQUENCE_HERE"
            app = ExtendedSequenceManipulator(example_seq, self.outfile)
            app.run()
            os.remove(self.outfile)


from scipy.cluster.hierarchy import dendrogram, linkage
from matplotlib import pyplot as plt
from scipy.spatial.distance import squareform
import pandas as pd
import warnings
from ..tidy_data.tidy_dendro import create_distance_matrix, get_clustered_sequences, label_bind_seqs, sort_binding_seqs




def levenshtein_dend(ax, sequencing_report, sample, batch_size,max_cluster_dist, font_settings, region_string):
    sample_report = sequencing_report[sequencing_report["Experiment"] == sample] ## insert test if sample not found
    report = sample_report.head(batch_size)
    aa = list(report[region_string])
    aa_clustered = get_clustered_sequences(aa, max_cluster_dist)
    # Create the distance matrix using the filtered list of sequences
    levenshtein_distance_matrix = create_distance_matrix(aa_clustered)

    # Convert to condensed form and create the dendrogram
    condensed_matrix = squareform(levenshtein_distance_matrix, checks=False)
    linked = linkage(condensed_matrix, 'single')
    dendrogram(linked,
                orientation='right',
                distance_sort='descending',
                show_leaf_counts=True,
                labels=aa_clustered,
                ax = ax
                )
    ax.set_xlabel("Levenshtein Distance", **font_settings)
    ax.set_ylabel("Sequences", **font_settings)
    title = "Levenshtein Distance between sequences in " + sample 
    ax.set_title(title,pad = 12, **font_settings)
    plt.tight_layout()


def dendo_binding(fig, sequencing_report,binding_data, sample,antigens, batch_size,max_cluster_dist,font_settings, region_string, ascending ):
    ax = fig.gca()

    sample_report = sequencing_report[sequencing_report["Experiment"] == sample] ## insert test if sample not found
    report = sample_report.head(batch_size)
    aa = report[region_string]
    aa = pd.DataFrame(aa)
    pref_columns = antigens + [region_string]
    b_data = binding_data[pref_columns]
    
    mix = pd.concat([aa, b_data])
    mix = mix.fillna(0)
    mix = mix.reset_index()
    
    aa_all = mix[region_string]
    aa_clustered = get_clustered_sequences(aa_all, max_cluster_dist)
    if len(aa_clustered) > 0:
        warnings.warn("More than 30 sequences with Levenshtein distance < " + str(max_cluster_dist) + " found. The resulting plot could be disordered. To change that please reduce the batch size, max_cluster_dist or you can adjust the string size of the sequences on the y axis with: (1) ax = plot.ax and (2) ax.tick_params(axis='y', labelsize=your_desired_size)")
    
    levenshtein_distance_matrix = create_distance_matrix(aa_clustered)
    condensed_matrix = squareform(levenshtein_distance_matrix, checks=False)
    linked = linkage(condensed_matrix, 'single')
    
    aa_clustered, binding_seqs, seq_val = label_bind_seqs(mix,region_string, aa_clustered, antigens, )
    binding_seqs_sorted, binding_values_sorted = sort_binding_seqs(binding_seqs, seq_val, ascending)


    dendrogram(linked,
            orientation='right',
            distance_sort='descending',
            show_leaf_counts=True,
            labels=aa_clustered,
             ax = ax
            )
    ax = plt.gca()
    ax.set_xlabel("Levenshtein Distance", **font_settings)
    ax.set_ylabel("Sequences", **font_settings)
    title = "Levenshtein Distance between sequences in " + sample
    
    ax.set_title(title,pad = 12, **font_settings)
    fig.show()
    fig2 = plt.figure(2)
    ax2 = fig2.gca()
    bars = ax2.barh(binding_seqs_sorted, binding_values_sorted)
    ax2.set_ylabel('Sequences with binding data', **font_settings)
    ax2.set_xlabel('Binding Value', **font_settings)
    fig2.tight_layout()
    fig.tight_layout()
    return fig2

    



import numpy as np
import editdistance
import networkx as nx

def cleaning(report, max_ld, min_ld, region_string):
    G = nx.Graph()
    aa = list(report[region_string])
    num_strings = len(aa)
    string_index = [*range(0, num_strings, 1)]
    max_ld = max_ld
    min_ld = min_ld
    for seq in aa:
        G.add_node(seq)
    for i in range(len(aa)):
        string1 = aa[i]
        index_1 = i
        for index in range(index_1):
            string2 = aa[index]
            distance = editdistance.distance(string1, string2)
            if distance>= min_ld and distance <= max_ld:
                G.add_edge(string1, string2)
    return G
